/* AlbumCover */
.albumcover-title-label {
    color: @theme_fg_color;
}

.albumcover-artist-label {
    font-size: smaller;
    color: @theme_fg_color;
}

/* AlbumsView */

#albums gridview {
  padding: 18px;
}

#albums gridview > child {
  margin: 9px;
  padding: 6px;
}

/* By default in Adwaita dark this is not transparent */
.tile {
    background-color: transparent;
}

.playlist-icon {
    padding: 10px;
    border-radius: 50%;
    background-color: alpha(@theme_fg_color, 0.1);
}

.disc-label {
    color: alpha(@theme_fg_color, 0.8);
    padding: 12px 0;
}

/* PlayerToolbar */

.smooth-scale {
    padding-top: 0px;
    padding-bottom: 0px;
}

/* We use background-image as a workaround on the StarImage widget to
   enable transitions between the non-starred and starred icons. */
.star {
    background-image: -gtk-icontheme('non-starred-symbolic');
    color: alpha(@theme_fg_color, 0.3);
    transition: 100ms linear;
}

.star:hover {
    color: alpha(@theme_fg_color, 0.6);
}

.star:selected {
    background-image: -gtk-icontheme('starred-symbolic');
    color: @theme_selected_bg_color;
    transition: 100ms linear;
}

.grey-image {
    color: alpha(@theme_fg_color, 0.3);
}

/* Used in AlbumWidget and ArtistAlbumsWidget */
.playing-song-label {
    font-weight: bold;
}

/* TwoLineTip */
.tooltip-title {
  font-weight: bold;
}

.playlistdialog-row:selected {
    color: @theme_fg_color;
    background-color: @theme_insensitive_bg_color;
}

.playlistdialog-row:selected label {
    color: @theme_text_color;
}

/* SearchView */
.search-header {
  font-weight: bold;
}

/* dnd icon for SongWidget */
.drag-handle {
  color: alpha(@theme_fg_color, 0.4);
}

.drag-handle:backdrop {
  color: alpha(@theme_unfocused_fg_color, 0.4);
}

/* Last.fm dialog */
.lastfm-status-label {
    font-weight: bold;
}

/* SongsView */
/* boxed-list style does not work for
GtkListView at the moment */
.songs-list {
    border-left: 1px solid @borders;
    border-right: 1px solid @borders;
    border-top: 1px solid @borders;
}

.songs-list:first-child {
    border-top-left-radius: 12px;
    border-top-right-radius: 12px;
}

.songs-list:last-child {
    border-bottom-left-radius: 12px;
    border-bottom-right-radius: 12px;
}

/* SongWidget and SongsView rows*/
.songwidget,
.songs-list row {
  padding: 12px;
}
